package EmailExtractor;
# This package is for reading MBox files and extracting individual emails
# into a database
use strict;
use Mail::MboxParser;
use Data::Dumper;
use Email::Find; #extract email addresses from email
use Date::Parse; #correct dates
use Date::Manip qw(UnixDate ParseDate);
use Digest::SHA1 qw(sha1_base64);
use File::Basename;
use POSIX qw(strftime); #print dates

our $seperator = "|";

sub new {
    my ($type,%options) = @_;
    my $type = shift;
    $options{seperator} ||= $seperator;
    return bless(\%options,$type);
}

sub csv_line {
    my $self = shift;
    return join( $self->{seperator}, (map { escape($_) } @_) ).$/;
}

sub header_string { 
    my $self = shift;
    return "#".$self->csv_line(@_);
}

#warning this mutates the email map
# obj, listname, handler
# listname is undef or blank stdin is assumed (filename)
# handler is a sub which takes (obj, valuehashref, emailhashref, errorarrayref, msg)
#   msg is a Mail::MboxParser::Message
# this is kind of a functional or block driven interface
sub parse_file {
    my ($self,$listname,$handler) = @_;
    die "Handler is not a coderef" unless ref($handler) eq 'CODE';
    $self->{email_map} = {} unless   $self->{email_map};
    my $email_map = $self->{email_map};
    my @errors = ();
    my $olistname = $listname;
    my $fd = \*STDIN;
    my $listname = $listname || "";
    my $filename = $listname;
    my $basename = basename($filename);
    warn "Parsing ".($listname||"STDIN");
    open($fd,"<",$listname) if $listname;
    my $mb = Mail::MboxParser->new( $fd, decode     => 'ALL' );
    warn "MBox Parser ready!";
    my $cnt = 0;
    while (my $msg = $mb->next_message) {
        # so we just parse each message and try to extract as much
        # as we can
        my $id = $msg->id;
        unless ($id) {
            my $error = "Message did not have an ID!$/".$msg->as_string;
            warn "Error: $error";
            push @errors, $error;
            $id = gen_id($msg);
        }
        my $body = $msg->body($msg->find_body);
        {
            # get the list name if it is a pipe from stdin
            if (!$olistname) {
                $listname = $msg->header->{"list-id"} || $listname;
            }
        }
        my $sig = $body->signature || "";

        my $subject = $msg->header->{subject};
        #warn "Subject: $subject";
        my $fname = $msg->from->{name};
        my $femail = $msg->from->{email};
        my ($date,$utime);
        eval {
            ($date,$utime) = get_date($msg);
        };
        if ($@) {
            my $error = "$@ : Utime undefined? [$utime] [$date] [$subject] [$femail] [$id]"; #.$msg->from_line.$/.Dumper($msg).$msg->as_string if !$utime;
            #my $error = "Could not calculate utime!$/".$msg->as_string;
            warn "Error: $error";
            push @errors, $error;
            next;
        }

        my $daterev = strftime "%Y-%m-%d", gmtime $utime;
        my $timerev = strftime "%H:%M:%S", gmtime $utime;
        my @tos = map { my @a = ();
                        push @a,$_->{name} if $_->{name};
                        push @a, $_->{email};
                        join(" ",@a) } $msg->to;
        my $to = join(", ", @tos);

        my @ccs = map { my @a = ();
                        push @a,$_->{name} if $_->{name};
                        push @a, $_->{email};
                        join(" ",@a) } $msg->cc;
        my $cc = join(", ", @ccs);



        my $nent = $msg->num_entities || 0;
        #this is so you can parse through things that weren't quoted 
        #and those that were
        my $quotes = $msg->body($msg->find_body)->quotes;
        my $quote0 = (exists $quotes->{0})?join($/,@{$quotes->{0}}):"";
        my $quote1 = (exists $quotes->{1})?join($/,@{$quotes->{1}}):"";
        my $quote2 = (exists $quotes->{2})?join($/,@{$quotes->{2}}):"";
        my $quoterest = join($/,
                             map { join($/,@{$quotes->{$_}}) } 
                             (sort {$a <=> $b} (grep { $_ > 2 }  (keys %$quotes))));
        my %emails = ();
        my $nemails = find_emails($body, sub { $emails{$_[0]->format} = $_[1];  }) || 0;
        my $emails = join(", ",sort(keys %emails));

        my %h = (
                 listname => $listname,
                 list => $listname,
                 id => $id,
                 date => $date,
                 daterev => $daterev,
                 timerev => $timerev,
                 utime => $utime,
                 fname => $fname,
                 from_name => $fname,
                 femail => $femail,
                 from_email => $femail,
                 to => $to,
                 cc => $cc,
                 nent => $nent,
                 entities => $nent,
                 subject => $subject,
                 body => $body,
                 quote0 => $quote0,
                 quote1 => $quote1,
                 quote2 => $quote2,
                 quoterest => $quoterest,
                 sig => $sig,
                 nemails => $nemails,
                 emails => $emails,
                 nbodyemails => $nemails,
                 bodyemails => $emails,
                 tos => \@tos,
                 ccs => \@tos,
                 emails_hash => \%emails,
                 ymd => ymd_from_utime($utime),
                 filename => $filename,
                 basename => $basename,
                 listname =>  $listname,
                 list_id => $msg->header->{"list-id"},
                );

        # now we go through all of our emails and associate them!
        $email_map->{$femail}->{$fname}++;
        foreach my $to (@tos) {
            my ($email,$name) = get_single_email($to);
            $email_map->{$email}->{$name}++;            
        }
        while (my ($email,$name) = each %emails) {
            $email_map->{$email}->{$name}++;
        }
        &$handler($self, \%h, $email_map, \@errors, $msg);
    }
    return @errors;
}

sub escape {
    my $sep = $seperator;
    my $v;
    my $self;
    if (@_ == 2) {
        ($self,$v) = @_;
        $sep = $self->{seperator};
    } else {
        $v = $_[0];
    }
    $v =~ s#\\#\\\\#gs;
    $v =~ s#\Q$seperator\E#\\$seperator#gs;
    $v =~ s#(\r|\n)# NEWLINE #gs;
    return $v;
}
sub get_single_email {
    my ($str) = @_;
    my $email = "";
    my $name = "";
    my $nemails = find_emails($str, 
                              sub { 
                                  $email = $_[0]->format || "";
                                  $name = $_[1] || "";
                              });
    my @o = ($email,$name);
    return @o;
}

#sub get_date {
#    my ($msg) = @_;
#    my $date = $msg->header->{date} || $msg->header->{received} || $msg->header->{"x-list-recei#ved-date"} || $msg->from_line;
#    if ($date =~ /From\s+[^\s]+\s+(.*)$/) {
#        $date = $1;
#    }
#    return $date;
#}

sub get_date {
    my ($msg) = @_;
    foreach my $date ($msg->header->{date}, $msg->header->{received}, $msg->header->{"x-list-received-date"} , $msg->from_line) {
        if ($date =~ /From\s+[^\s]+\s+(.*)$/) {
            $date = $1;
        }
        my $utime = UnixDate($date,"%s");
        $utime = $utime || str2time($date);
        if ($utime) {
            my @out = ($date,$utime);
            return @out;
        }
    }
    die "Date Failure!";
}

sub get_utime_from_date {
    my ($date) = @_;
    my $utime = UnixDate($date,"%s");
    $utime = $utime || str2time($date);
    return ($utime);
}




sub random_id {
    my @a = ('a'..'z','A'..'Z');
    return join("",(map { $a[rand(@a)] }  (1..26)));
}
sub gen_id {
    my ($msg) = @_;
    return sha1_base64($msg->as_string);
}
# $msg is a Mail::MboxParser::Mail

sub email_map { 
    $_[0]->{email_map};
}
# or preferably

sub ymd_from_utime {
    return strftime "%Y-%m-%d", gmtime($_[0]);
}

1;
